/* Copyright OCHIAI Laboratory 
modified by Noriko Imafuji in 1998, Nobuko KOMAGATA in 1999, 
Chihiro Nakai in 2002 and Mari Yasuda in 2003
*/
#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <math.h>
#define  EOD   555
#define  NMAX  20
#define  MAX1  300  /* crossing point * 2  */
#define  MAX2  600
#define  MAX32 9600 /* note that MAX1*2*16 */
#define  DEPTH 0.2
typedef struct prdata {
    int 	a,b,c;
} PDATA;
typedef struct point{
  int x, y;
} POINT;

typedef struct cpinfo {
  int	h;	/* head */
  int	t;	/* tail */
} CPINFO;

extern jmp_buf		env;
static double		eps = 1.0e-18;
static double 		PI = 3.14159265358979;
static int     		g_n = 0, g_n2 = 0;
static int			g_com, g_nump[NMAX];
static PDATA   		g_prd[MAX1];
static int 			g_rightedge[MAX2][2];
static int     		g_edgepoint[MAX1][MAX2];
static CPINFO		g_cpinfo[MAX2];
static double		*g_KGM, *g_INV, *g_KP, *g_KP2;
double				*g_KPRO[3];

static int			g_n3, g_n31;
static int			g_neout,g_neout2,g_neout21,g_outd,g_nd,g_ned[MAX1];
static int 			NextPoint(int, int, PDATA*, int*);
static int			BeforePoint(register int, PDATA *, int *);
static void 		FindRight(int, PDATA*, int*);
int					GetDrawData(FILE *, int *, int *);
static void 		OverCross(void);
static int 			Euler(void);
static void 		OuterDomain(void);
static void 		MakeMatrix(void);
static void 		Swap(double*, double*);
static int 			InverseMatrix(void);
static int 			CalKGM(void);
static int 			CalKP(void);
static void 		NormalKP(void);
static int 			DisplacementMatrix(void);
static double		Norm_2D(int, int); //modified
static double		Norm_3D(int, int); //additional
static int 			Mmod(int, int, int);
static void			ConnectSmoothly_2D(void); //additional
static void			ConnectSmoothly_3D(void); //additional
static void			ConnectSmoothly_3D_Shifted(void); //additional
static int 			Smoothing(void);
static int 			GetMemory(void);
static void 		ReleaseMemory(void);
int					Sign(register int); //additional
void				CompletePrd(int *, int, int, int, int *, PDATA *); //additional
int					ReadPrd(FILE *, int *, int *, int *, int *, PDATA *); //additional
void				NextUpDown(int, int, int); //additional
void				SetUp3DCoord(void); //additional
void				FreeDrawMemory(void); //additional
static void OverCross(void)
{
    register int	i,j,k,s;
    int				over[MAX2];
    for (k = 1, i = 1; i <= g_n2; i++)
    if (g_prd[i].b % 2 != 0) over[i] = k;
    else over[i] = k++;
    i = g_n2;
    if (g_prd[g_n2].b % 2 != 0)
    while (over[i] == k)
    over[i--] = 1;
    for (k = 1, i = 1; i <= g_n2; i++)
		if (g_prd[i].b % 2 == 0)
			g_cpinfo[i].t = over[i] + g_n2; /* under point */
		else
			g_cpinfo[i].t = over[g_prd[i].a] + g_n2; /* over point */
    for (s = 0, i = 1; i <= g_com; i++) {
        g_cpinfo[s+1].h = g_cpinfo[s+g_nump[i]].t;
        for (j = s+2; j <= s+g_nump[i]; j++)
			g_cpinfo[j].h = g_cpinfo[j-1].t;
        s += g_nump[i];
    }
}
static int Euler(void)
{
    register int 	i,k,i1;
    POINT	 		bnddomain[MAX2];
    for (i = 0; i <= g_n2; i++)
    bnddomain[i].x = bnddomain[i].y = 0;
    for (g_nd = 0, i = -g_n2; i <= g_n2; i++) {
        if ((i < 0 && bnddomain[-i].x == 0) ||
        (i > 0 && bnddomain[i].y == 0)) {
            g_nd++; k = 0; i1 = i;
            do {
                k++;
                g_edgepoint[g_nd][k] = i1;
                if (i1 > 0)	bnddomain[i1].y = g_nd;
                else bnddomain[-i1].x = g_nd;
                i1 = (i1 > 0) ? g_rightedge[i1][1]: g_rightedge[-i1][0];
            } while ((i1 < 0 && bnddomain[-i1].x != g_nd) ||
            (i1 > 0 && bnddomain[i1].y != g_nd));
            g_ned[g_nd] = k;
        }
    }
    return(g_n-g_n2+g_nd);
}
static void OuterDomain(void)
{
    register int i,ld,neld;
    
    ld = 1; neld = 0;
    for (i = 1; i <= g_nd; i++) {
        if (g_ned[i] > neld) {
            ld = i;
            neld = g_ned[i];
        }
    }
    g_neout = neld;
    g_neout2 = g_neout + g_neout;
    g_outd = ld;
    g_neout21 = g_neout2 + 1;
}
static void MakeMatrix(void)
{
    register int 	i,j,k;
    int 			e1,e2,c1,c2,v1,v2,ed, abs(int);
    double 			cod,cod2;
    
	for (i = 1; i <= g_n3; i++) { /* BugFix 2003/2/26 */	
		for (j = 1; j <= g_n3; j++)
			g_KGM[j+i*g_n31] = 0.0;
		for (j = 1; j <= g_neout2; j++) 
			g_KP[j+i*g_neout21] = 0.0;
	}    
    for (i = 1; i <= g_neout; i++) {
        ed = g_edgepoint[g_outd][i];
        v1 = abs(ed);
        v2 = (ed > 0) ? g_cpinfo[ed].t: g_cpinfo[-ed].h;
        g_KP[2*i-1+v1*g_neout21] = 0.25;
        g_KP[2*i+v2*g_neout21] = 0.08333333333333;
    }
    for (i = 1; i <= g_nd; i++) {
        if (i != g_outd) {
            if (g_ned[i] == 2)
            cod = 0.04166666666666 / 1000.0;
            else
            cod = 0.04166666666666 / g_ned[i];
            cod2 = cod + cod;
            for (j = 1; j <= g_ned[i]; j++) {
                ed = g_edgepoint[i][j];
                e1 = abs(ed);
                c1 = (ed > 0) ? g_cpinfo[ed].t:
                g_cpinfo[-ed].h;
                for (k = 1; k <= g_ned[i]; k++) {
                    ed = g_edgepoint[i][k];
                    e2 = abs(ed);
                    c2 = (ed > 0) ? g_cpinfo[ed].t: g_cpinfo[-ed].h;
                    g_KGM[e2+e1*g_n31] = cod2;
                    g_KGM[c2+e1*g_n31] = cod2;
                    g_KGM[e2+c1*g_n31] = cod;
                    g_KGM[c2+c1*g_n31] = cod;
                }
            }
        }
    }
    for (i = 1; i <= g_n2; i++) {
        c1 = g_cpinfo[i].t;
        c2 = g_cpinfo[i].h;
        g_KGM[c1+i*g_n31] += 0.33333333333333;
        g_KGM[c2+i*g_n31] += 0.33333333333333;
        g_KGM[i+c1*g_n31] += 0.16666666666666;
        g_KGM[i+c2*g_n31] += 0.16666666666666;
    }
}
void Swap(double *x, double *y)
{
    double z;
    z = *x; *x = *y; *y = z;
}
static int InverseMatrix(void)
{
    register int	i, j, k;
    double  		maxentry, sign, value, *ptr, fabs(double);
    register double pivot, TEMP;
    register int	pos, column, index;
    
    for (i = 1; i <= g_n3; i++) {
        pos = i; pos *= g_n31;
        for (j = 1; j <= g_n3; j++)
        g_INV[pos+j] = (i == j) ? 1.0 : 0.0;
    }       /* set unit matrix at g_INV[] */
    value = sign = 1.0;
    k = 1;
    while (k <= g_n3) {
        index = column = k * g_n31; /* set column index */
        maxentry = fabs(g_KGM[index+k]);
        for (i = k+1; i <= g_n3; i++) {
            pos = i; pos *= g_n31;
            if (fabs(g_KGM[pos+k]) > maxentry) {
                column = pos;
                maxentry = fabs(g_KGM[pos+k]);
            }
        }
        if (maxentry <= eps) {
            printf("Displacement matrix is singular.");
            longjmp(env, 1);
            return 0;
        }
        if (column != index) {
            for (j = 1; j <= g_n3; j++) {
                Swap(g_KGM+index+j,g_KGM+column+j);
                Swap(g_INV+index+j,g_INV+column+j);
            }
            sign = -sign;
        }
        pivot = g_KGM[index+k];
        value *= pivot;
        for (j = 1; j <= g_n3; j++) {
            g_KGM[index+j] /= (double)pivot;
            g_INV[index+j] /= (double)pivot;
        }
        for (i = 1; i <= g_n3; i++) {
            pos = i; pos *= g_n31;
            TEMP = g_KGM[pos+k];
            if (i != k)
            for (j = 1; j <= g_n3; j++) {
                g_KGM[pos+j] -= TEMP * g_KGM[index+j];
                g_INV[pos+j] -= TEMP * g_INV[index+j];
            }
        }
        k++;
    }
    ptr = g_KGM; g_KGM = g_INV; g_INV = ptr;
    if (fabs(value) <= eps) {
        printf("Displacement matrix is singular.");
        longjmp(env, 1);
        return 0;
    }
    else return 1;
}
static int CalKGM(void)
{
    register int 	i,j;
    int				size;
    for (i = 1; i <= g_n3; i++)
    for (j = 1; j <= g_n3; j++) {
        size = i; size *= g_n31; size += j;
        g_KGM[size] = (i==j) ? 1.0-g_KGM[size]: -g_KGM[size];
    }
    if (InverseMatrix() == 1) return 1;
    else return -1;
}
static int CalKP(void)
{
    register int 	i,j,k,i31,i32,i33;
    register double s1,s2,s3,kj,*ptr;
    register int	size;
    double			fabs(double);
    
    for (i = 1; i <= g_n; i++) {
        i33 = 3 * i; i32 = i33 - 1; i31 = i33 - 2;
        for (j = 1; j <= g_neout2; j++) {
            s1 = s2 = s3 = 0.0;
            for (k = 1; k <= g_n3; k++) {
                kj = g_KP[j+k*g_neout21];
                if (fabs(kj) > eps) {
                    size = i31; size *= g_n31; size += k;
                    s1 += g_KGM[size] * kj;
                    size = i32; size *= g_n31; size += k;
                    s2 += g_KGM[size] * kj;
                    size = i33; size *= g_n31; size += k;
                    s3 += g_KGM[size] * kj;
                }
            }
            g_KP2[j+i31*g_neout21] = s1;
            g_KP2[j+i32*g_neout21] = s2;
            g_KP2[j+i33*g_neout21] = s3;
        }
    }
    ptr = g_KP; g_KP = g_KP2; g_KP2 = ptr;
    return 1;
}
static void NormalKP(void)
{
    register int 	i,j;
    register double s;
    for (i = 1; i <= g_n3; i++) {
        for (s = 0.0, j = 1; j <= g_neout2; j++)
        s += g_KP[j+i*g_neout21];
        for (j = 1; j <= g_neout2; j++)
        g_KP[j+i*g_neout21] /= s;
    }
}
static int DisplacementMatrix(void)
{
    if (CalKGM() == -1) return -1;
    if (CalKP() == -1) return -1;
    NormalKP();
    return 1;
}
static double Norm_2D(int i, int j)
{
    register double x,y;
    double			sqrt(double);
    x = g_KPRO[0][i] - g_KPRO[0][j];
    y = g_KPRO[1][i] - g_KPRO[1][j];
    return sqrt(x*x+y*y);
}
static double Norm_3D(int i, int j)
{
    register double x,y,z;
    double			sqrt(double);
    x = g_KPRO[0][i] - g_KPRO[0][j];
    y = g_KPRO[1][i] - g_KPRO[1][j];
    z = g_KPRO[2][i] - g_KPRO[2][j];
    return sqrt(x*x+y*y+z*z);
}
static int Mmod(int j, int x, int y)
{
    register int i, k;
    for (k = 0, i = 2; i <= j; i++)
    k += g_nump[i-1];
    return (x <= y) ? x: 16*k+x%y;
}
static void ConnectSmoothly_2D(void)
{
    register int 	i,j,k,z,t;
    int 			i1,i2,i3,i4,q,nn,mm,nump16,q1;
    double 			nor1,nor2,nor3,alpha;
    for (t = 0, j = 1; j <= g_com; j++) {
        q = (t + 1) * 16;
        nn = g_nump[j];
        t += nn; nump16 = t * 16; q1 = 16;
        for ( k = 1; k <= 3; k++) {
            q1 /= 2; nn *= 2;
            i2 = q - q1; i3 = i2 + q1; i4 = i3 + q1;
            nor2 = Norm_2D(i2,i3); nor3 = Norm_2D(i3,i4);
            for (i = 1; i <= nn; i++) {
                i1 = i2; i2 = i3; i3 = i4;
                i4 = Mmod(j,i4+q1,nump16);
                nor1 = nor2; nor2 = nor3;
                nor3 = Norm_2D(i3,i4);
                mm = Mmod(j,i3-q1/2,nump16);
                alpha = (16.0 - 5.0 / k)/nor2;
                for (z = 0; z < 2; z++) {
                    g_KPRO[z][mm] =
                    (g_KPRO[z][i2]+g_KPRO[z][i3])/2
                    +((g_KPRO[z][i2]-g_KPRO[z][i1])/nor1
                    +(g_KPRO[z][i3]-g_KPRO[z][i4])/nor3)/alpha;
                }
            }
        }
    }
}
static void ConnectSmoothly_3D(void)
{
    register int 	i,j,k,z,t;
    int 			i1,i2,i3,i4,q,nn,mm,nump16,q1;
    double 			nor1,nor2,nor3,alpha;
    for (t = 0, j = 1; j <= g_com; j++) {
        q = (t + 1) * 16;
        nn = g_nump[j];
        t += nn; nump16 = t * 16; q1 = 16;
        for ( k = 1; k <= 3; k++) {
            q1 /= 2; nn *= 2;
            i2 = q - q1; i3 = i2 + q1; i4 = i3 + q1;
            nor2 = Norm_3D(i2,i3); nor3 = Norm_3D(i3,i4);
            for (i = 1; i <= nn; i++) {
                i1 = i2; i2 = i3; i3 = i4;
                i4 = Mmod(j,i4+q1,nump16);
                nor1 = nor2; nor2 = nor3;
                nor3 = Norm_3D(i3,i4);
                mm = Mmod(j,i3-q1/2,nump16);
                alpha = (16.0 - 1.0 / k)/nor2;
                for (z = 0; z <=2; z++) {
                    g_KPRO[z][mm] =
                    (g_KPRO[z][i2]+g_KPRO[z][i3])/2
                    +((g_KPRO[z][i2]-g_KPRO[z][i1])/nor1
                    +(g_KPRO[z][i3]-g_KPRO[z][i4])/nor3)/alpha;
                }
            }
        }
    }
}
static void ConnectSmoothly_3D_Shifted(void)
{
    register int 	i,j,k,z,t;
    int 			i1,i2,i3,i4,q,nn,mm,nump16,q1;
    double 			nor1,nor2,nor3,alpha;
    for (t = 0, j = 1; j <= g_com; j++) {
        q = (t + 1) * 16;
        nn = g_nump[j];
        t += nn; nump16 = t * 16; q1 = 16;
        for ( k = 1; k <= 3; k++) {
            q1 /= 2; nn *= 2;
            i2= q - q1 / 2; i3 = i2 + q1; i4 = i3 + q1;
            nor2 = Norm_3D(i2,i3); nor3 = Norm_3D(i3,i4);
            for (i = 1; i <= nn; i++) {
                i1 = i2; i2 = i3; i3 = i4;
                i4 = Mmod(j,i4+q1,nump16);
                nor1 = nor2; nor2 = nor3;
                nor3 = Norm_3D(i3,i4);
                mm = Mmod(j, i2+q1/2, nump16);
                alpha = (16.0 - 1.0 / k)/nor2;
                for (z = 0; z <=2; z++) {
                    g_KPRO[z][mm] =
                    (g_KPRO[z][i2]+g_KPRO[z][i3])/2
                    +((g_KPRO[z][i2]-g_KPRO[z][i1])/nor1
                    +(g_KPRO[z][i3]-g_KPRO[z][i4])/nor3)/alpha;
                }
            }
        }
    }
}
static int Smoothing(void)
{
    register int 	i,j,z, size;
    double 			s,cout[MAX1][2], sin(double), cos(double);
    double			scale[2];
    register double *KPRO0;

    scale[0] = -1; scale[1] = 1;
    size = 2 * sizeof(double) * g_n31;
    KPRO0 = (double *)malloc(size);
    if (KPRO0 == NULL) {
        printf("Memory Exhausted on Code 202");
        return -1;
    }
    for (i = 1; i <= g_neout2; i++) {
        s = PI / g_neout * (i + 1);
        cout[i][0] = sin(s); cout[i][1] = cos(s);
    }
    for (z = 0; z < 2; z++) {
        for (i = 1; i <= g_n3; i++) {
            s = 0.0;
            for (j = 1; j <= g_neout2; j++)
            s += g_KP[j+i*g_neout21] * cout[j][z];
            KPRO0[z+(i-1)*2] = s;
        }
    }
    for (i = 1; i <= g_n2; i++) {
        for (z = 0; z < 2; z++) {
            g_KPRO[z][16*i-8] = KPRO0[z+(i-1)*2] * scale[z];
            g_KPRO[z][16*i] = KPRO0[z+(g_cpinfo[i].t-1)*2] * scale[z];
        }
    }
    free((char*)KPRO0);
    ConnectSmoothly_2D();
    return 1;
}
int NextPoint(register int i, int n, PDATA *prd, int *nump)
{
    if ((prd[i].c == prd[i+1].c) && (i < n))
    return(i+1);
    else
    return(i+1-nump[prd[i].c]);
}
int BeforePoint(register int i, PDATA *prd, int *nump)
{
    if ((1 < i) && (prd[i].c == prd[i-1].c))
    return(i-1);
    else
    return(i-1+nump[prd[i].c]);
}
static int GetMemory(void)
{
    int		size, dsize;
    dsize = sizeof(double);
    size = dsize * g_n31 * g_n31;
    g_KGM = (double *)malloc(size);
    if (g_KGM==NULL) {
        printf("Memory Exhausted on Code 200.1");
        return -1;
    }
    size = dsize * g_n31 * g_n31;
    g_INV = (double *)malloc(size);
    if (g_INV == NULL) {
        printf("Memory Exhausted on Code 200.2");
        return -1;
    }
    size = dsize * g_n31 * g_neout21;
    g_KP = (double *)malloc(size);
    if (g_KP==NULL) {
        printf("Memory Exhausted on Code 201.1");
        return -1;
    }
    size = dsize * g_n31 * g_neout21;
    g_KP2 = (double *)malloc(size);
    if (g_KP2 == NULL) {
        printf("Memory Exhausted on Code 201.2");
        return -1;
    }
    return 1;
}
static void ReleaseMemory(void)
{
    free((char*)g_KP); g_KP = NULL;
    free((char*)g_KP2); g_KP2 = NULL;
    free((char*)g_KGM); g_KGM = NULL;
    free((char*)g_INV); g_INV = NULL;
}
void FindRight(int n2, PDATA *prd, int *nump)
{
    register int	i, k;
    for (i = 1; i <= n2; i++)
    if ((prd[i].b == 1) || (prd[i].b == 4))
    g_rightedge[i][1] = NextPoint(prd[i].a,n2,prd,nump);
    else
    g_rightedge[i][1] = -prd[i].a;
    for (i = 1; i <= n2; i++) {
        k = g_rightedge[BeforePoint(i,prd,nump)][1];
        if (k > 0)
        g_rightedge[i][0] = -BeforePoint(k,prd,nump);
        else
        g_rightedge[i][0] = NextPoint(-k,n2,prd,nump);
    }
}
int Sign(register int x)
{
    return(x > 0) ? 1: -1;
}
void
CompletePrd(int *pd, int n, int n2, int com, int *nump, PDATA *prd)
{ // [f[^S[f[^ɕϊ
    int		i, j, k, abs(int);
    
    for (i = 1; i <= n; i++) {
        prd[i].b = prd[n+i].b = 0;
    }
    for (i = 1; i <= n; i++) // _̌̎di[
    prd[abs(pd[i])].b = Sign(pd[i]) + 3;
    /* _̕ */
    /* 2: }CiX, 4: vX */
    for (k = 0, i = 1; i <= n2; i++) {
        if (prd[i].b == 0) { // _̌̎di[
            k++;
            prd[i].a = abs(pd[k]);
            prd[i].b = Sign(pd[k]) + 2; /* _̕ */
            prd[abs(pd[k])].a = i;      /* 1: }CiX, 3: vX */
        }
    }
    for (k = 0,i = 1; i <= com; i++) { // ݖڂ̐ԍ
        for (j = 1; j <= nump[i]; j++) prd[++k].c = i;
    }
}
int ReadPrd(FILE *fp_r, int *n, int *n2, int *com, int *nump, PDATA *prd)
{ // [f[^t@Cǂݍ
    int		i, s;
    int		pd[MAX1/2];
    fscanf(fp_r, "%d", n2);
    if ((*n2 <= 0) || (*n2 == EOD)) return(EOD);
    fscanf(fp_r, "%d", com);
    if (abs(*com) >= NMAX) return(0);
    *n = *n2 / 2; nump[0] = 0;
    for (i = 1; i <= *com; i++) fscanf(fp_r,"%d",nump+i);
    for (s = 0, i = 1; i <= *com; i++) s += nump[i];
    if (*n2 != s) return(0);
    for (i = 1; i <= *n; i++) {
        fscanf(fp_r,"%d", pd+i);
        if (abs(pd[i]) > *n2) return(0);
    }
    CompletePrd(pd, *n, *n2, *com, nump, prd);
    return(1);
}
void NextUpDown(int s, int t, int j)
{
    int		i, k, jt, st, st1, np;
    if (s == j) {
        st = 16 * j + 1; jt = t;
    }
    else {
        st = 16 * j; jt = j;
    }
    st1 = 16 * (j + 1);
    np = NextPoint(jt, g_n2, g_prd, g_nump);
    if (g_prd[jt].b % 2 && g_prd[np].b % 2)
    for (i = st; i <= st1; i++)
    g_KPRO[2][i] = DEPTH;
    else if (g_prd[jt].b % 2 && !(g_prd[np].b % 2))
    for (k = 0, i = st; i <= st1; i++, k++)
    g_KPRO[2][i] = DEPTH - 2.0 * k * DEPTH / 16;
    else if (!(g_prd[jt].b % 2) && g_prd[np].b % 2)
    for (k = 0, i = st; i <= st1; i++, k++)
    g_KPRO[2][i] = -DEPTH + 2.0 * k * DEPTH / 16;
    else if (!(g_prd[jt].b % 2) && !(g_prd[np].b % 2))
    for (i = st; i <= st1; i++)
    g_KPRO[2][i] = -DEPTH;
}
void SetUp3DCoord(void)
{
    int		i, j, s, t;
    for (s = 0, i = 1; i <= g_com; i++) {
        t = s + g_nump[i];
        for (j = s; j <= t; j++)
        NextUpDown(s, t, j);
        s += g_nump[i];
    }
}
void FreeDrawMemory(void)
{
    if (g_KPRO[0] != NULL) {
        free((char*)g_KPRO[0]);
        g_KPRO[0] = NULL;
    }
    if (g_KPRO[1] != NULL) {
        free((char*)g_KPRO[1]);
        g_KPRO[1] = NULL;
    }
    if (g_KPRO[2] != NULL) {
        free((char*)g_KPRO[2]);
        g_KPRO[2] = NULL;
    }
}
int GetDrawData(FILE *fp_r, int *com, int *nump)
{
    int		i, j, size, code;
    code = ReadPrd(fp_r, &g_n, &g_n2, &g_com, g_nump, g_prd);
    if (code == EOD) return(EOD);
    else if (!code) return(0);
    if (g_n2 > 300) {
        printf("The lenght is too large.");
        longjmp(env, 1);
        return -1;
    }
    else if (g_n2 <= 2 && g_com == 1) {
        printf("This knot is trivial.");
        return -1;
    }
    else if (g_n2 == 0 && g_com > 1) {
        printf("A trivial component is obtained.");
        return -1;
    }
    *com = g_com;
    for (i = 1; i <= g_com; i++)
    nump[i] = g_nump[i];
    FreeDrawMemory();
    OverCross();
    FindRight(g_n2,g_prd,g_nump);
    if (Euler() == 2) {
        g_n3 = 3 * g_n; g_n31 = g_n3 + 1;
        OuterDomain();
        if (GetMemory() == -1) {
            ReleaseMemory();
            printf("Memory Exhausted on Code 203");
            longjmp(env, 1);
            return -1;
        }
        MakeMatrix();
        if (DisplacementMatrix() != 1) {
            ReleaseMemory();
            longjmp(env, 1);
            return -1;
        }
        size = sizeof(double);
        size *= MAX32;
        for (i = 0; i < 3; i++) {
            g_KPRO[i] = (double *)malloc(size);
            if (g_KPRO[i] == NULL) {
                printf("Memory Exhausted on Code 204");
                for (j = 0; j < i; j++) {
                    free((char*)g_KPRO[j]);
                    g_KPRO[j] = NULL;
                }
                longjmp(env, 1);
                return -1;
            }
        }
        if (Smoothing() == -1) {
            ReleaseMemory();
            longjmp(env, 1);
            return -1;
        }
        SetUp3DCoord();
        for(i=0; i<20; i++) {
            ConnectSmoothly_3D();
            ConnectSmoothly_3D_Shifted();
        }
        ReleaseMemory();
        return 1;
    }
    else {
        printf("The Euler number is not 2. May be, this link is splittable.");
        return -1;
    }
}
