/* Copyright OCHIAI LAB. of Nara Women's University
programmed by K. Kambe, T. Taninaka, and M. Ochiai
2004/1/25
This program uses an algorithm in 
"AN O(m log n)-TIME ALGORITHM FOR THE MAXIMAL PLANAR SUBGRAPH PROBLEM"
H.Chai, K. Han, and R. Tarjan 
SIAM J. COMPUT. vol. 22 No. 6, pp 1142-1162, 1993
Note that this program is not complete and so we recommend you to use another 
program by Hopcroft-Tarjan algorithm. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <math.h>
#include <cstdlib> 

#define maxv   100
#define maxv2  202
#define maxe   300
#define ev1    200

#define True   1
#define False  0

#define	max(x,y)	((x > y) ? x: y)
#define	min(x,y)	((x < y) ? x: y)

#define minpre(x,y)	   ((pre[x]<pre[y]) ? x: y)
#define maxpre(x,y)	   ((pre[x]<pre[y]) ? y: x)
#define low1(x) 		edge[x].low1
#define low2(x) 		edge[x].low2
#define head(x) 		edge[x].head
#define tail(x) 		edge[x].tail
#define att(x)			edge[x].att

typedef struct link { /* vertex link */
		int    vtx;
		struct link *next;
} link;

typedef struct link2 { /* edge link */
		int    lt, rt;
		struct link2 *next;
} link2;

typedef struct link3 { /* only use in lplane */
		struct link2 *lchild,*rchild;
		struct link3 *next;
} link3;

typedef struct eedge {
		int head,tail;
		int lw1,lw2;	/* low1, low2 of original graph */
		int low1,low2;	/* low1, low2 of deformed graph */
		link3 *att;
} eedge;

static int		g_nv, g_ne, g_k, g_ct, nonplanar, totalplanar;
static link		*free_top=NULL; 
static link2	*free_top2=NULL, *est=NULL, *bst=NULL; 
static link3	*free_top3=NULL;
static int		pre[maxv]; /* preorder tree search ordering */
static int		lpt1[maxv], lpt2[maxv];
static link		*a[maxv], *pa[maxv];
static link2	*philist[maxv2]; /* use in only palist */

/* only use in lplane */
static eedge	edge[maxe];
static int		edgenumber;

static jmp_buf	env;
static int		retcode = 0;

extern FILE			*infile;
extern char			filename[]; /* *filename no good */
extern int			INDATA;

static FILE			*outfile = NULL;

//char	OUTFILE[17] = "CaiHanTarjan.out";
//char	INFILE[11] = "INDATA.dat";

void	readgraph(), write_illegal(), write_planar(), write_nonplanar(),
		palist(),  dfs(),  init(), releasememory(),  releasememory2(), releasememory3();
void	judge();

link	*new_link(), *pop(),*push();
link2	*new_link2(), *push2(), *pop2();
link3	*new_link3(), *push3(), *pop3();

link *new_link(void)
{
	link	*ptr;

	ptr = (link *)malloc(sizeof(link));
	if (ptr == NULL) {
		printf("Memory exhausted.\n");
		exit(1);
	}
	else return(ptr);
}

link2 *new_link2(void)
{
	link2  *ptr;

	ptr = (link2 *)malloc(sizeof(link2));
	if (ptr == NULL) {
		printf("Memory exhausted.\n");
		exit(1);
	}
	else return(ptr);
}

link3 *new_link3(void)
{
	link3  *ptr;

	ptr = (link3 *)malloc(sizeof(link3));
	if (ptr == NULL) {
		printf("Memory exhausted.\n");
		exit(1);
	}
	else return(ptr);
}

void readgraph(void)
{
	int		i, vtx, edge, nu;
	link	*temp, *current;

	fprintf(outfile, "*** Adjacency List ***\n");

	fscanf(infile, "%d", &nu);
	fscanf(infile, "%d", &vtx);  
	fprintf(outfile, "\n %3d:", vtx);
	temp = new_link();
	a[vtx] = current = temp;
	g_nv = 1; 
	g_ne = 0;

	while (fscanf(infile, "%d", &edge) != EOF) {
		if (edge == 0) {
			fscanf(infile, "%d", &vtx);
			current->vtx = 0;
			current->next = NULL;
			if(vtx == 0) {
				g_ne = g_ne / 2;
				break;
			}
			else {
				++g_nv;
				fprintf(outfile, "\n %3d:", vtx);
				temp = new_link();
				a[vtx] = current = temp;
			}
		}
		else {
			fprintf(outfile, "%6d", edge);
			if (vtx == edge) 
				write_illegal();
			else {
				++g_ne;
				temp = new_link();
				current->vtx = edge;
				current->next = temp;
				current = temp;
			}
		}
	}
	fprintf(outfile, "\n");
	
	for (i = 0; i <= g_nv; i++) {
		pre[i] = 0; pa[i] = NULL;
	}

	fclose(infile);
	infile = NULL;
}

void write_illegal(void) 
{
	printf("\n ??? illegal data ??? \n");
	retcode = -1;
	longjmp(env, 1);
}

link *push( int head, link *tail)
{

	link	*ptr;

	if (free_top == NULL) 
		ptr = new_link();
	else {
		ptr = free_top;
		free_top = free_top->next;
	}
	ptr->vtx = head;
	ptr->next = tail;
	return (ptr);
}

link2 *pop2(link2 *dl2)
{
	link2  *ptr;

	ptr = dl2->next;
	dl2->next = free_top2;
	free_top2 = dl2;
	return(ptr);
}

link3 *pop3(link3 *dl2)
{
	link3  *ptr;

	ptr = dl2->next;
	dl2->next = free_top3;
	free_top3 = dl2;
	return(ptr);
}

void palist(int v, int w)
{
	int    j, vx, wx, phimax, phi,tail;
	link2  *ptr, *ptr2;
	int    count = 0;

	if (lpt1[w] > pre[v]) 
		est = pop2(est);
	else {
		phimax = 2 * pre[v];
		while (pre[est->lt] >= pre[w]) {
			vx = est->lt; wx = est->rt;
			fprintf(outfile, "%6d -%3d", vx, wx);
			++count;
			if (count % 6 == 0) 
				fprintf(outfile, "\n");
			if (pre[vx] > pre[wx]) 
				phi = 2 * pre[wx];	/* leaf edge */	
			else {					/* tree edge */
				if (lpt2[wx] >= pre[vx]) 
					phi = 2 * lpt1[wx]; 
				else 
					phi = 2 * lpt1[wx] + 1;
			}
			phimax = max(phimax, phi);
			ptr = est;
			est = ptr->next;
			ptr->next = philist[phi];
			philist[phi] = ptr;
		}
		phi = 2 * lpt1[w];
		ptr = est;
		est = ptr->next;
		ptr->next = philist[phi];
		philist[phi] = ptr;
		for (j = phimax; j >= 2*pre[v]; --j) {
			if (philist[j] != NULL) {
				ptr = philist[j];
				ptr2 = ptr->next;
				tail = ptr->lt;
				pa[tail] = push(ptr->rt,pa[tail]);
				/* make a properly ordered adjacency list */
				while (ptr2 != NULL) {
					ptr = ptr2;
					tail = ptr->lt;
					pa[tail] = push(ptr->rt,pa[tail]);
					ptr2 = ptr2->next;
				}
				ptr->next = free_top2;
				free_top2 = philist[j];
				philist[j] = NULL;
			}
		}
	}
	fprintf(outfile, "%6d -%3d\n", v, w);
}

link2 *push2(int left, int right, link2 *top)
{
	link2   *ptr;

	if (free_top2 == NULL) 
		ptr = new_link2();
	else {
		ptr = free_top2;
		free_top2 = free_top2->next;
	}
	ptr->lt = left; ptr->rt = right;
	ptr->next = top;
	return (ptr);
}

link3 *push3(link2 *lch,link2 *rch,link3 *top)
{
	link3   *ptr;

	if (free_top3 == NULL) 
		ptr = new_link3();
	else {
		ptr = free_top3;
		free_top3 = free_top3->next;
	}
	ptr->lchild = lch; ptr->rchild = rch;
	ptr->next = top;
	return (ptr);
}

link *pop(link *dl)
{
	link	*ptr;

	ptr = dl->next;
	dl->next = free_top;
	free_top = dl;
	return(ptr);
}

int number(int a,int b)
{
	int j;

	for( j=1;j<edgenumber;j++) {
		if(edge[j].head!=a) 
			continue;
		if(edge[j].tail!=b) 
			continue;	
		return(j);
	}
	return(0);
}

link *successorstk(int a,link *nextv)
{
	link	*rev,*nxt,*lst,*top=NULL;

	while(nextv!=NULL) {
		top = push(number(a,nextv->vtx), top);
		nextv = nextv->next;
	}
	lst=top;
	/*	reverse */
	for (rev = NULL; lst != NULL; ) {
		nxt = rev;
		rev = lst;
		lst =lst->next;
		rev->next = nxt;
	}
	return(rev);
}

link3 *mcopy(link3 *r)
{
	link3	*rnew,*rnext,*rrev,*rlst,*rnxt;
	link2	*nw[2],*first[2],*rev,*lst,*nxt;
	int		i;

	rnew=NULL;
	for(rnext=r;rnext!=NULL;rnext=rnext->next) {
		first[0] = rnext->lchild;
		first[1] = rnext->rchild;
		for (i=0;i<2;i++) {
			nw[i] = NULL;
			for(nxt=first[i];nxt!=NULL;nxt=nxt->next) 
				nw[i] = push2(nxt->lt,nxt->rt,nw[i]);
			lst = nw[i];
			for(rev=NULL;lst!=NULL;) {
				nxt = rev;
				rev = lst;
				lst =lst->next;
				rev->next = nxt;
			}
			nw[i]=rev;
		}
		rnew = push3(nw[0],nw[1],rnew);
	}
	rlst=rnew;
	for(rrev=NULL;rlst!=NULL;) {
		rnxt = rrev;
		rrev = rlst;
		rlst =rlst->next;
		rrev->next = rnxt;
	}
	return(rrev);
}

int normal1(int lw1,int a,link2 *XY)
{
	link2	*XYnext;

	for(XYnext=XY;XYnext!=NULL;XYnext=XYnext->next) {
		if(pre[lw1] < pre[XYnext->rt] && pre[XYnext->rt] < pre[a]) 
			return(1); /** normal **/
	}
	return( 0); /** not normal **/
}

void low12set(int x,int *lw1,int *lw2)
{
	if(pre[*lw1] > pre[x]) {
		*lw2 = *lw1;
		*lw1 = x;
	}
	else {
		if(x!=*lw1) {
			if(pre[x] < pre[*lw2]) 
				*lw2 = x;
		}
	}
}

link2  *merge(link2 *X,link2 *B,int *lw1,int *lw2)
{
	link2	*Bnext,*Xnext,*Btmp, *lst, *rev, *nxt;

	if(X==NULL) 
		return(B);
	if (B==NULL) 
		return(X);

	Bnext=B;
	Btmp=NULL;
	Xnext=X;
	while(Xnext!=NULL) {
		if (Bnext!=NULL && pre[Xnext->rt]<=pre[Bnext->rt]) {
			Btmp=push2(Bnext->lt,Bnext->rt,Btmp);
			Bnext=pop2(Bnext);
		}
		else {
			Btmp=push2(Xnext->lt,Xnext->rt,Btmp);
			Xnext=Xnext->next;
		}
	}
	while(Bnext!=NULL) {
		Btmp=push2(Bnext->lt,Bnext->rt,Btmp);
//		low12set(Bnext->rt,lw1,lw2);
		Bnext=pop2(Bnext);
	}
	lst = Btmp; /* BugFix 2004/2/3 by M. Ochiai */
	for(rev=NULL; lst!=NULL; ) { /* reverse link2 */
		nxt = rev;
		rev = lst;
		lst = lst->next;
		rev->next = nxt;
	}
	return(rev);
}

int interlace(int blow1,link2 *xy)
{	 
	if(xy==NULL) 
		return(0); /* no interlace */
	if(pre[blow1] >= pre[xy->rt]) 
		return(0); /* no interlace */
	return(1); /* interlace */
}

link3 *occurence(link3 *r, int b)
{
	link2	*X,*X1;

	X = r->lchild;
	X1 = NULL;
	while(X != NULL) {
		if(X->rt != b) 
			X1 = push2(X->lt,X->rt,X1);
		X = pop2(X);
	}
	while(X1 != NULL) {
		X = push2(X1->lt,X1->rt,X);
		X1 = pop2(X1);
	}
	r->lchild = X;
	X = r->rchild;
	X1 = NULL;
	while(X != NULL) {
		if(X->rt != b) 
			X1 = push2(X->lt,X->rt,X1);
		X = pop2(X);
	}
	while(X1 != NULL) {
		X = push2(X1->lt,X1->rt,X);
		X1 = pop2(X1);
	}
	r->rchild = X;
	if(r->lchild == NULL) {
		if(X == NULL) 
			r = pop3(r);
		else {
			r->rchild = r->lchild;
			r->lchild = X;
		}
	}
	else {
		X1 = r->lchild;
		if(X != NULL && pre[X1->rt] < pre[X->rt]) {
			r->rchild = r->lchild;
			r->lchild = X;
		}
	}
	return(r);
}

int seth12(link3 *r,int lr)
{
	link3	*tmp;
	link2	*X;
	int h;
/** lr ---- lchild	  -- h1	
	lr ---- rchild	  -- h2 
**/
	h = 0;
	switch(lr) {
	case 1: 
		for(tmp=r;tmp!=NULL;tmp=tmp->next) {
			if(tmp->lchild!=NULL) {
				X = tmp->lchild;
				return(X->rt);
			}
		}
		break;
	case 2:
		for(tmp=r;tmp!=NULL;tmp=tmp->next) {
			if(tmp->rchild!=NULL) {
				X = tmp->rchild;
				return(X->rt);
			}
		}
		break;	
	}
	return(0);
}

void del(int a,int b,int delv,int *lo1,int *lo2)
{
	link	*X1;

	/**  leaf edge **/
	X1 = NULL;
	if(pre[a]>pre[b]) {
		while(pa[a]!=NULL) {
			if(pa[a]->vtx!=delv)  
				X1 = push(pa[a]->vtx,X1);			
			pa[a] = pop(pa[a]);
		}
		while(X1!=NULL) {
			pa[a] = push(X1->vtx,pa[a]);
			X1 = pop(X1);
		}
		*lo1 = g_nv+1;
		*lo2 = g_nv+1;
	}
	/** tree edge **/
	else {
		while(pa[b]!=NULL) {
			del(b,pa[b]->vtx,delv,&low1(number(b,pa[b]->vtx)),&low2(number(b,pa[b]->vtx)));
			low12set(low1(number(b,pa[b]->vtx)),lo1,lo2);
		}
	}
}

void printrr(int i)
{
	link3	*tmp,*p;
	link2	*X,*Y,*Xtmp,*Ytmp,*px,*py;
	int		phi1,phi2;

	p = NULL;
	for( tmp=edge[i].att;tmp!=NULL;tmp=tmp->next) {
		X = tmp->lchild;
		Y = tmp->rchild;
		px = NULL;
		py = NULL;
		for(Xtmp=X;Xtmp!=NULL;Xtmp=Xtmp->next) 
			px = push2(Xtmp->lt,Xtmp->rt,px); 
		for(Ytmp=Y;Ytmp!=NULL;Ytmp=Ytmp->next) 
			py = push2(Ytmp->lt,Ytmp->rt,py); 
		p = push3(px,py,p);
	}

	if (edge[i].head > edge[i].tail) 
		phi1 = phi2 = 2 * edge[i].tail;	/* leaf edge */	
	else {					/* tree edge */
		if (edge[i].low2 >= edge[i].head) {
			phi1 = 2 * edge[i].lw1;
			phi2 = 2 * edge[i].low1;
		}
		else {
			phi1 = 2 * edge[i].lw1 + 1;
			phi2 = 2 * edge[i].low1 + 1;
		}
	}

	fprintf(outfile,"edge=%d(%d-%d),low1=%d->%d,low2=%d->%d, phi=%d->%d",i,head(i),tail(i),edge[i].lw1,low1(i),edge[i].lw2,low2(i),phi1,phi2);
	fprintf(outfile,"\n        att(e)=[ ");
	while( p!=NULL) {
		X = p->lchild;
		Y = p->rchild;
		fprintf(outfile," [[ ");
		while(X!=NULL) {
			fprintf(outfile,"%d-%d ",X->lt,X->rt); 
			X = pop2(X);
		}
		fprintf(outfile,"],[ ");
		while(Y!=NULL) {
			fprintf(outfile,"%d-%d ",Y->lt,Y->rt); 
			Y = pop2(Y);
		}
		fprintf(outfile,"]] ");
		p = pop3(p);
	}
	fprintf(outfile," ]\n");
}

void lplanar (int a,int b,int enumber,int l) 
{
	link2	*Bi,*X,*Y,*X1;
	link	*succes,*stmp;
	link3	*r,*nextr,*tmp;
	int		low1e1,low2e1,e1,ei,ej,delv,i,j,low1e2,low1ei,low2ei;
	int		h1,h2,h3,min1,min2;

	if(pre[a]> pre[b]) { /* this edge is leaf */
		att(enumber) = push3(push2(a,b,NULL),NULL,NULL);
		return;
	}
	if(pa[b]==NULL) {
		att(enumber) = NULL;
		return;
	}
	/*	step (1)	*/
	/* organize e1,....en into a heap  */
	succes=successorstk(b,pa[b]); 
	/* about e1,e2,,,,ek (successor of e)*/
	/* depend on increasing order of phi(ei) */
	e1 = succes->vtx;
	succes = pop(succes);

	lplanar(head(e1),tail(e1),e1,l);	/* initialize by att(e1)  */

	r = mcopy(att(e1));
	low1e1 = low1(e1);
	low2e1 = low2(e1);
	r = occurence(r,b);	 /* occurrence(to b)  delete */
	j = 0;
	ej = e1;
	i = 1;
	while(succes!=NULL) {
		ei = succes->vtx;
		if(pre[low1(ei)]>=pre[ b]) {

			lplanar(head(ei),tail(ei),ei,b);

			succes = pop(succes);
		}
		/** compute h1,h2 and h3 **/
		else {
			h1 = seth12(r,1);
			h2 = seth12(r,2);
			h3 = 0;
			for(tmp=r;tmp!=NULL;tmp=tmp->next) {
				if(tmp->lchild==NULL) 
					continue;
				if(normal1(low1e1,l,tmp->lchild)) {
					X = tmp->lchild;
					h3 = X->rt;
					break;
				}
			}
			for(tmp=r;tmp!=NULL;tmp=tmp->next) {
				if(tmp->rchild==NULL) 
					continue;
				if(normal1(low1e1,l,tmp->lchild)) {
					X = tmp->rchild;
					h3 = maxpre(h3,X->rt);
					break;
				}
			}
			/**	 condition AA **/
			if((j==0&&pre[low1(ei)]<pre[h2])||(pre[low1(ej)]<pre[low1(ei)] && 
				pre[low1(ei)]<pre[h2])||(pre[low1(ej)]<pre[low1(ei)] && 
				pre[low1(ei)]<minpre(h3,l))) {
				delv = low1(ei);
				del(head(ei),tail(ei),delv,&low1(ei),&low2(ei));
				if(pre[head(ei)]>pre[tail(ei)]) 
					succes = pop(succes);
				nonplanar = True;

				/* temporary code to avoid an infinite loop */
				printf("\n *** delete edge ***\n");
				printf("%3d --%3d\n", head(ei),tail(ei));
				fprintf(outfile,"\n *** delete edge ***\n");
				fprintf(outfile, "%6d --%3d", head(ei),tail(ei));
				write_nonplanar();
				longjmp(env, 1);
				/*------------------------------------------*/
			}

			/** conditon BB  **/
			else if((pre[low2(ei)]<pre[b ] && pre[low1(ei)]<pre[h2]) 
					||(pre[low2(ei)]<pre[b]&& pre[low1(e1)]<pre[low1(ei)] 
					&& pre[low1(ei)]<minpre(h3,l))||(pre[low2(ei)]<pre[l] && 
					pre[low1(ei)]<pre[h3])) {
				delv = low2(ei);
				del(head(ei),tail(ei),delv,&low1(ei),&low2(ei));
				nonplanar = True;
				
				/* temporary code to avoid an infinite loop */
				printf("\n *** delete edge ***\n");
				printf("%3d --%3d\n", head(ei),tail(ei));
				fprintf(outfile,"\n *** delete edge ***\n");
				fprintf(outfile, "%6d --%3d", head(ei),tail(ei));
				write_nonplanar();
				longjmp(env, 1);
				/*------------------------------------------*/
			}
			/*	  step (2)	 */
			else {
				succes=pop(succes);

				lplanar(head(ei),tail(ei),ei,b);
				
				/** this is Bi[i] **/
				nextr = mcopy(edge[ei].att);		
				low1ei = low1(ei);
				low2ei = low2(ei);
				/** occurrence(to b)  delete **/ 
				nextr = occurence(nextr,b);
				Bi = NULL;
				do {
					X = nextr->lchild;
					Y = nextr->rchild;
					if(normal1(low1(ei),b,X) && normal1(low1(ei),b,Y)) {
						nonplanar = True;

						write_nonplanar();
						longjmp(env, 1);  /** not plannar **/
					} 
					else { 
						Bi=merge(X,Bi,&low1ei,&low2ei);
						Bi=merge(Y,Bi,&low1ei,&low2ei);
					}
					nextr = nextr->next;
				} while(nextr!=NULL);
				/*	 step (3)    */
				if(i==1) {
					/* set low(e2) -- this is used in step (3) */
					low1e2 = low1(ei);
					for(stmp=succes;stmp!=NULL;stmp=stmp->next) {
						if(pre[low1e2] > pre[low1(stmp->vtx)]) 
							low1e2 = low1(stmp->vtx);
					}
					while(r!=NULL) {
						tmp = r->next;
						if(tmp==NULL) 
							break;
						X = tmp->lchild;
						if(pre[X->rt]<=pre[low1e2]) 
							break;
						tmp->lchild = merge(X,r->lchild,&low1e1,&low2e1);
						tmp->rchild = merge(tmp->rchild,r->rchild,&low1e1,&low2e1);
						r = pop3(r);
					}
				}
				/*   step (4)   */
				if(Bi!=NULL) {
					if(interlace(low1ei,r->lchild)==1) {
						if(interlace(low1ei,r->rchild)==0) {
							/****** interlace X only *********/
							r->rchild = merge(Bi,r->rchild,&low1e1,&low2e1);
							X = r->lchild;
							Y = r->rchild;
							if(Y!=NULL && pre[X->rt]<pre[Y->rt]) {
								X1 = r->lchild;
								r->lchild = r->rchild;
								r->rchild = X1;
							}
						} 
						else {
							nonplanar = True;
							write_nonplanar();
							longjmp(env, 1);
						}
					} 
					else {	 /** add [B[j],[]] to last block **/
						r=push3(Bi,NULL,r);
						min1 = minpre(low1e1,low1ei);
						if(maxpre(low1e1,low1ei)!=min1) 
							min2 = maxpre(low1e1,low1ei);
						else {
							if(minpre(low2e1,low2ei)!=min1) 
								min2 = minpre(low2e1,low2ei);
							else 
								min2 = maxpre(low2e1,low2ei);
						}
						low1e1 = min1;
						low2e1 = min2;	
					}
					j = i;
					ej = ei;
				}
				i++;
			}
		}
	}
	low1(enumber)=low1e1;	/* modified low1, low2 */
	low2(enumber)=low2e1;
	att(enumber)=r;
	return;
}

void swap(int *x, int *y)
{
	int	temp;
	temp = *x; *x = *y; *y = temp;
}

void edgelist(int v)
{
	link *nextp;

	for(nextp=pa[v];nextp!=NULL;nextp=nextp->next) {
		edge[edgenumber].head = v;
		edge[edgenumber].tail = nextp->vtx;;
		if(pre[v]>pre[nextp->vtx]) { /* set edge[].low1, edge[].low2 */
			edge[edgenumber].low1 = edge[edgenumber].lw1=nextp->vtx;
			edge[edgenumber].low2 = edge[edgenumber].lw2=g_nv+1;
			edgenumber++;
		} 
		else {
			edge[edgenumber].low1 = edge[edgenumber].lw1=lpt1[nextp->vtx];
			edge[edgenumber].low2 = edge[edgenumber].lw2=lpt2[nextp->vtx];
			edgenumber++;
			edgelist(nextp->vtx);
		}
	}
}

void printpalist(int v)
{
	link	*tmp,*tmp1;

	fprintf(outfile,"vertex- %d  successor-> ",v);
	tmp1 = NULL;
	for(tmp=pa[v];tmp!=NULL;tmp=tmp->next) 
		fprintf(outfile," %d ",tmp->vtx);
	fprintf(outfile,"\n");
	for(tmp=pa[v];tmp!=NULL;tmp=tmp->next) {
		if(pre[v]>pre[tmp->vtx]) 
			continue;
		printpalist(tmp->vtx);
	}
}

void dfs(int v, int u)
{
	int		w,na,count;
	int		nowe;

	pre[v] = lpt1[v] = lpt2[v] = ++g_k;
	while (a[v]->next != NULL) {
		w = a[v]->vtx;
		if (pre[w] == 0) {
			est = push2(v, w, est); 			
			dfs(w, v);
			if (lpt1[w] < lpt1[v]) { 
				lpt2[v] = min(lpt1[v], lpt2[w]);
				lpt1[v] = lpt1[w];
			}
			else if (lpt1[w] == lpt1[v]) 
				lpt2[v] = min(lpt2[v], lpt2[w]);
			else
				lpt2[v] = min(lpt2[v], lpt1[w]);
			if (lpt1[w] >= pre[v]) { 
				fprintf(outfile,"\n ### No.(%3d)biconnected component ###\n\n",++g_ct);
				palist(v, w);
				nonplanar = False;
				if(pa[v]!=NULL) {
					nowe = edgenumber;

					edgelist(v); /* set edge[].low1, edge[].low2 */

					lplanar(0,v,0,0);

					fprintf(outfile," ** successor list **\n");
					printpalist(v);
					fprintf(outfile,"\n *** att(e) ***\n");
					for(na=nowe;na<edgenumber;na++) {
						if(low1(na)<g_nv+1)  
							printrr(na);
					}
					fprintf(outfile,"\n *** maximal planar subgraph ***\n");
					count = 0;
					for(na=nowe;na<edgenumber;na++) {
						if(low1(na)<g_nv+1) {
							fprintf(outfile, "%6d -%3d", head(na),tail(na));
							++count;
							if (count % 6 == 0) 
								fprintf(outfile, "\n");
						}
					}
					if(nonplanar) {
						count = 0;
						fprintf(outfile,"\n *** delete edge (of this subgraph) ***\n");
						for(na=nowe;na<edgenumber;na++) {
							if(low1(na)>=g_nv+1) {
								fprintf(outfile, "%6d -%3d", head(na),tail(na));
								++count;
								if (count % 6 == 0) 
									fprintf(outfile, "\n");
							}
						}
					}
				}
				/* embed(v); */
				if (nonplanar == True) {
					totalplanar=True;
					fprintf(outfile,"\n\n --> This biconnected component is nonplanar.\n\n"); 
					return;
				} 
				else 
					fprintf(outfile,"\n\n --> This biconnected component is planar.\n\n");
				if (pa[v] != NULL) 
					pa[v] = pop(pa[v]);
			}
		}
		else if ((w != u) && (pre[w] < pre[v])) {
			est = push2(v, w, est); 		       
			if (pre[w] < lpt1[v]) {
				lpt2[v] = lpt1[v];
				lpt1[v] = pre[w];
			} 
			else if (pre[w] > lpt1[v]) 
				lpt2[v] = min(lpt2[v], pre[w]);     
		}
		a[v] = pop(a[v]);
	}
}

void write_planar(void)
{
	fprintf(outfile, "\n *** This graph is planar by Cai-Han-Tarjan algorithm. *** \n");
	retcode = 1;
}

void write_nonplanar(void)
{
	fprintf(outfile, "\n *** This graph is nonplanar by Cai-Han-Tarjan algorithm. *** \n");
	totalplanar = True;
	retcode = 0;
}

void init(void)
{ 
	int		i;

	free_top = NULL; 
	free_top2 = est = NULL;
	free_top3 = NULL;
	bst = new_link2();
	bst->next = NULL;
	bst->lt = bst->rt = 0;	   
	totalplanar = False;
	for (i = 0; i <= 2*g_nv+2; ++i) 
		philist[i] = NULL;
	g_k = g_ct = 0;
	edgenumber = 1;
}

void releasememory(void)
{
	link	*p;
	link2	*p2;
	link3	*p3;

	while (free_top != NULL) {
		p = free_top;
		free_top = free_top->next;
		free(p);
	}
	while (free_top2 != NULL) {
		p2 = free_top2;
		free_top2 = free_top2->next;
		free(p2);
	}
	while (free_top3 != NULL) {
		p3 = free_top3;
		free_top3 = free_top3->next;
		free(p3);
	}
}

void judge(void)
{
	int		i;

	if (g_nv < 5 || g_ne < 9) {
		write_planar();
	}
	else if ((g_ne > 3*g_nv-3) || (g_nv >= 3 && g_ne > 3*g_nv-6)) {
		write_nonplanar();
	}
	else {
		init();
		for (i = 1; i <= g_nv; ++i) {
			if (pre[i] == 0) { /* to check each connected component */
				if(a[i]->next == NULL) {/* to avoid isolated vertices */
					++g_ct;
					fprintf(outfile,"\n ### No.(%3d)biconnected component ### \n\n",g_ct);
					fprintf(outfile, "%d(isolated)\n\n", i);
					fprintf(outfile, "--> This component is planar. \n\n");
				} 
				else 
					dfs(i, 0);
			}
		}
		if (totalplanar == False) {
			write_planar();
		}
		if (totalplanar == True) {
			write_nonplanar();
		}
	} 
	fprintf(outfile, "preorder tree search ordering. \n"); 
	for (i = 1; i <= g_nv; ++i) 
		fprintf(outfile, "%d  ", pre[i]);
	fprintf(outfile, "\n");
	releasememory();
}

int open_outfile(void)
{
	if ((outfile = fopen("CaiHanTarjan.out", "w")) == NULL) {
		printf("Can't open file : %s\n", "CaiHanTarjan.out");
		return(0);
	}
	return(1);
}

int CaiHanTarjan(void)
{
	int		code;

	if (infile != NULL) {
		; /* already opened */
	}
	else if ((infile = fopen("INDATA.dat", "r")) == NULL)  {
		printf("Can't open file : %s\n", "INDATA.dat");
		return(-1);
	}

	outfile = NULL;
	code = open_outfile();
	if (code == 0)
		return(-1);
	if (INDATA == 0)
		fprintf(outfile, "*** Testing for the data file %s *** \n", filename);

	readgraph();

	if (setjmp(env)) {
		goto ext;
	}

	judge();

ext:
	fclose(outfile);
	fflush(stdout);
	return(retcode);
}
