//*****************************************//
// Ochiai Lab. Nara Women's University     //
// Programmed by Chihiro Taninaka          //
// All rights reserved. March 1999         //
//*****************************************//

#ifndef __MAKE_GRAPH_H__
#define __MAKE_GRAPH_H__

#include <iostream>
#include <string>
using namespace std;

#include <math.h>
#include "..\CGraph\CGraph.h"
#include "..\Wada\draw.h"
#include "..\Tools\Matrix.h"

class CMakeGraph {
	int Vnum;	//_̐
	int *mtx;
	double (*xy)[2];
public:
	CMakeGraph(int v){ 
		Vnum=v;
		mtx = new int[v*v];
		xy  = new double[v][2];
		for(int i=0; i < v; i++) {
			xy[i][0] = cos(i*2.*pi/v+pi/2);
			xy[i][1] = sin(i*2.*pi/v+pi/2);
		}
		for(i = 0; i < v*v; i++)  mtx[i]=0;
	}

	CMakeGraph(int v, iMatrix &mat)
	{
		Vnum = v; mtx = mat.to_array();
		xy  = new double[v][2];
		for(int i=0; i < v; i++) {
			xy[i][0] = cos(i*2.*pi/v+pi/2);
			xy[i][1] = sin(i*2.*pi/v+pi/2);
		}
	}

	CMakeGraph(iostream fin)
	{
		CGraph g(fin);
		if(g.NumV() != CGraph::G_EMPTY) {
			Vnum = g.NumV();
			iMatrix mat = g.GetAdjMat();
			mtx = mat.to_array();
			xy  = new double[Vnum][2];
			for(int i=0; i < Vnum; i++) {
				xy[i][0] = cos(i*2.*pi/Vnum+pi/2);
				xy[i][1] = sin(i*2.*pi/Vnum+pi/2);
			}			
		} else {
			Vnum = 0; mtx = NULL; xy = NULL;
		}
	}

	~CMakeGraph()
	{ delete[] mtx; }

	void ChangeIncidence(int v1, int v2)
	{
		if(v1 == v2) return;
		if(mtx[v1*Vnum+v2]) mtx[v1*Vnum+v2]=mtx[v2*Vnum+v1] = 0;
		else				mtx[v1*Vnum+v2]=mtx[v2*Vnum+v1] = 1;
	}

	int SearchVertex(double x, double y)
	{
		double dis=0.1, min = 0.1 , d;
		int no=-1;
		for(int i = 0; i < Vnum; i++) {
			d = (xy[i][0]-x)*(xy[i][0]-x)+(xy[i][1]-y)*(xy[i][1]-y);
			if(d < min) {
				min = d; no = i;
			}
		}
		return no;
	}

	void DrawGraph(draw &d)
	{
		int i;
		double r = d.Pixel()* 2;
		char buf[256];

		string str;
		for(i = 0; i < Vnum; i++) {
//			d.DrawCircle(xy[i][0],xy[i][1],r,0,rgb(0,0,1));
			d.PaintDisc (xy[i][0],xy[i][1],r,rgb(0,0,1));
			sprintf(buf, "%d", i+1);
			str = buf;
			d.DrawString(xy[i][0]*1.1, xy[i][1]*1.1, str, 0.1, "Times");
			for(int j = i; j < Vnum; j++) {
				if(mtx[i*Vnum+j])
					d.DrawLine(xy[i][0], xy[i][1], xy[j][0], xy[j][1]);
			}
		}
	}
	void CheckVertex(draw &d, int vtx)
	{
		double r = d.Pixel()* 2;
		if(vtx < 0 || vtx >= Vnum) return;
			d.PaintDisc(xy[vtx][0],xy[vtx][1],r,rgb(1,0,0));
	}

	void Output(iostream fout)
	{
		fout << Vnum;
		for(int i=0; i<Vnum; i++) {
			fout << endl << i+1 << " ";
			for(int j=0; j<Vnum ; j++) {
				if(mtx[i*Vnum+j])
					fout << j+1 << " ";
			}
			fout << 0;
		}
		fout << 0 << endl;
	}
};
#endif

