// draw.h   1998 by Masaaki Wada. All rights reserved.
// 19980521 Masaaki Wada
// 19980614MW name change (mIDraw), bug fix (wedge)

#ifndef		_draw_
#define		_draw_

#include	"iDraw.h"
#include	"pvect.h"

#ifndef		pi
#define		pi	3.14159265358979323846
#endif


// ================================================================================
// draw

class draw {

// ---------- data members ----------
protected:
	iDraw&			mIDraw;
	double			mXMin;
	double			mXMax;
	double			mYMin;
	double			mYMax;
	int				mKeepRatio;
	double			mXPixel;
	double			mYPixel;
	double			mPixel;
	double			mXOrigin;
	double			mYOrigin;

// ---------- constructor and accessors ----------
public:
	draw(	iDraw& d,
			double xmin = -1.1, double xmax = 1.1,
			double ymin = -1.1, double ymax = 1.1,
			int keep_ratio = true );

	iDraw&			IDraw( void ) { return mIDraw; }
	double			XMin( void ) { return mXMin; }
	double			XMax( void ) { return mXMax; }
	double			YMin( void ) { return mYMin; }
	double			YMax( void ) { return mYMax; }
	double			XPixel( void ) { return mXPixel; }
	double			YPixel( void ) { return mYPixel; }
	double			Pixel( void ) { return mPixel; }

	void			SetRange( double xmin, double xmax, double ymin, double ymax )
						{
							mXMin = xmin; mXMax = xmax;
							mYMin = ymin; mYMax = ymax;
							Adjust();
						}
	void			KeepRatio( int keep_ratio ) { mKeepRatio = keep_ratio; Adjust(); }

	iRect			Frame( void ) { return mIDraw.Frame(); }
	int				Left( void ) { return mIDraw.Left(); }
	int				Right( void ) { return mIDraw.Right(); }
	int				Top( void ) { return mIDraw.Top(); }
	int				Bottom( void ) { return mIDraw.Bottom(); }
	void			SetFrame( iRect fr ) { mIDraw.SetFrame( fr ); Adjust(); }
	void			SetFrame( int l, int t, int r, int b )
						{ mIDraw.SetFrame( l, t, r, b ); Adjust(); }
protected:
	void			Adjust( void );   // called when mIDraw/range changed

// ---------- conversion ----------
public:
	double			X( int i ) { return ( mXOrigin + i * mXPixel ); }
	double			Y( int j ) { return ( mYOrigin - j * mYPixel ); }
	pvect			V( iPoint p ) { return V( p.i, p.j ); }
	pvect			V( int i, int j ) { return pvect( X( i ), Y( j ) ); }
	int				I( double x ) { return (int)( ( x - mXOrigin ) / mXPixel + 0.5 ); }
	int				J( double y ) { return (int)( ( mYOrigin - y ) / mYPixel + 0.5 ); }
	iPoint			P( pvect v ) { return iPoint( I( v.x ), J( v.y ) ); }
	iPoint			P( double x, double y ) { return iPoint( I( x ), J( y ) ); }

	int				Itrunc( double x ) { return (int)( ( x - mXOrigin ) / mXPixel ); }
	int				Jtrunc( double y ) { return (int)( ( mYOrigin - y ) / mYPixel ); }
	iPoint			Ptrunc( pvect v ) { return iPoint( Itrunc( v.x ), Jtrunc( v.y ) ); }
	iPoint			Ptrunc( double x, double y )
						{ return iPoint( Itrunc( x ), Jtrunc( y ) ); }

// ---------- drawing ----------
public:
// line
	void			DrawLine( pvect v0, pvect v1, double width = 0.0, rgb color = black );

	void			DrawLine( double x0, double y0, double x1, double y1,
													double width = 0.0, rgb color = black )
						{ DrawLine( pvect( x0, y0 ), pvect( x1, y1 ), width, color ); }

// cubic Bezier curve
	void			DrawBezier( pvect v0, pvect v1, pvect v2, pvect v3,
													double width = 0.0, rgb color = black );

	void			DrawBezier( double x0, double y0, double x1, double y1,
								double x2, double y2, double x3, double y3,
													double width = 0.0, rgb color = black )
						{ DrawBezier( pvect( x0, y0 ), pvect( x1, y1 ),
										pvect( x2, y2 ), pvect( x3, y3 ), width, color ); }

// circle
	void			DrawCircle( pvect center, double radius,
													double width = 0.0, rgb color = black )
						{ DrawArc( center.x, center.y, radius, 0, 360, width, color ); }

	void			DrawCircle( double x, double y, double radius,
													double width = 0.0, rgb color = black )
						{ DrawArc( x, y, radius, 0, 360, width, color ); }

// arc
	void			DrawArc( pvect center, double radius, int start, int sweep,
													double width = 0.0, rgb color = black );

	void			DrawArc( double x, double y, double radius, int start, int sweep,
													double width = 0.0, rgb color = black )
						{ DrawArc( pvect( x, y ), radius, start, sweep, width, color ); }

// disc
	void			PaintDisc( pvect center, double radius, rgb color = black )
						{ PaintWedge( center.x, center.y, radius, 0, 360, color ); }

	void			PaintDisc( double x, double y, double radius, rgb color = black )
						{ PaintWedge( x, y, radius, 0, 360, color ); }

// wedge
	void			PaintWedge( pvect center, double radius, int start, int sweep,
																		rgb color = black );

	void			PaintWedge( double x, double y, double radius, int start, int sweep,
																		rgb color = black )
						{ PaintWedge( pvect( x, y ), radius, start, sweep, color ); }

// polygon
	void			DrawPolygon( pvect vertex[], int num_vertices, rgb color = black );

	void			PaintPolygon( pvect vertex[], int num_vertices, rgb color = black );

// character string
	void			DrawString( pvect v, string str,
								int font_size = 10, string font_name = "",
								int centering = false, rgb color = black );

	void			DrawString( double x, double y, string str,
								int font_size = 10, string font_name = "",
								int centering = false, rgb color = black )
						{ DrawString( pvect( x, y ), str,
										font_size, font_name, centering, color ); }

	void			DrawString( pvect v, string str,
								double font_sz, string font_name = "",
								int centering = false, rgb color = black );

	void			DrawString( double x, double y, string str,
								double font_sz, string font_name = "",
								int centering = false, rgb color = black )
						{ DrawString( pvect( x, y ), str,
										font_sz, font_name, centering, color ); }
};   // end of class draw



// ================================================================================
// shapes

// --------------------------------------------------------------------------------
// shape : base class
class shape {
protected:
	rgb				mColor;
	pvect			mBase;
public:
	shape( rgb color = black, pvect p = pvect( 0.0, 0.0 ) )
											: mColor( color ), mBase( p ) {}
	shape( rgb color, double x, double y ) : mColor( color ), mBase( x, y ) {}
	virtual ~shape() {}

	rgb				Color( void ) { return mColor; }
	virtual void	SetColor( rgb color ) { mColor = color; }
	pvect			Base( void ) { return mBase; }
	virtual void	SetBase( pvect p ) { mBase = p; }

	virtual shape*	Clone( void ) { return new shape( *this ); }
	virtual void	Draw( draw& d ) {}
	virtual void	Translate( pvect v ) { mBase += v; }
	shape&			operator +=( pvect v ) { Translate( v ); return *this; }
};

inline shape&		operator +( shape s, pvect v ) { return ( s += v ); }

inline draw&		operator <<( draw& d, shape& s ) { s.Draw( d ); return d; }


// --------------------------------------------------------------------------------
// paint_out
class paint_out : public shape {
public:
	paint_out( rgb color = white ) : shape( color ) {}
	virtual shape*	Clone( void ) { return new paint_out( *this ); }
	virtual void	Draw( draw& d ) { d.IDraw() << iPaintOut( mColor ); }
	paint_out&		operator +=( pvect v ) { Translate( v ); return *this; }
};

inline paint_out&	operator +( paint_out s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// line : line segment
class line : public shape {
public:
	pvect			mP0;
	pvect			mP1;
	double			mWidth;

	line( pvect p0, pvect p1, double width = 0.0, rgb color = black )
						: shape( color ), mP0( p0 ), mP1( p1 ), mWidth( width ) {}
	line( double x0, double y0, double x1, double y1,
								double width = 0.0, rgb color = black )
						: shape( color ), mP0( x0, y0 ), mP1( x1, y1 ), mWidth( width ) {}

	virtual shape*	Clone( void ) { return new line( *this ); }
	virtual void	Draw( draw& d )
						{ d.DrawLine( mBase + mP0, mBase + mP1, mWidth, mColor ); }
	line&			operator +=( pvect v ) { Translate( v ); return *this; }
};

inline line&		operator +( line s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// bezier : cubic Bezier curve
class bezier : public shape {
public:
	pvect			mP0, mP1, mP2, mP3;
	double			mWidth;

	bezier( pvect p0, pvect p1, pvect p2, pvect p3, double width = 0.0, rgb color = black )
		: shape( color ), mP0( p0 ), mP1( p1 ), mP2( p2 ), mP3( p3 ), mWidth( width ) {}

	bezier( double x0, double y0, double x1, double y1,
			double x2, double y2, double x3, double y3,
			double width = 0.0, rgb color = black )
		: shape( color ), mP0( x0, y0 ), mP1( x1, y1 ), mP2( x2, y2 ), mP3( x3, y3 ),
																	mWidth( width ) {}

	virtual shape*	Clone( void ) { return new bezier( *this ); }
	virtual void	Draw( draw& d )
						{ d.DrawBezier( mBase + mP0, mBase + mP1,
										mBase + mP2, mBase + mP3, mWidth, mColor ); }
	bezier&			operator +=( pvect v ) { Translate( v ); return *this; }
};

inline bezier&		operator +( bezier s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// arc
class arc : public shape {
public:
	double			mRadius;
	int				mStart;
	int				mSweep;
	double			mWidth;

	arc( pvect center, double radius,
						int start, int sweep, double width = 0.0, rgb color = black )
		: shape( color, center ), mRadius( radius ),
						mStart( start ), mSweep( sweep ), mWidth( width ) {}
	arc( double x, double y, double radius,
						int start, int sweep, double width = 0.0, rgb color = black )
		: shape( color, x, y ), mRadius( radius ),
						mStart( start ), mSweep( sweep ), mWidth( width ) {}

	virtual shape*	Clone( void ) { return new arc( *this ); }
	virtual void	Draw( draw& d )
						{ d.DrawArc( mBase, mRadius, mStart, mSweep, mWidth, mColor ); }
	arc&			operator +=( pvect v ) { Translate( v ); return *this; }
};

inline arc&			operator +( arc s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// circle
class circle : public arc {
public:
	circle( pvect center = pvect( 0.0, 0.0 ), double radius = 1.0,
											double width = 0.0, rgb color = black )
		: arc( center, radius, 0, 360, width, color ) {}
	circle( double x, double y, double radius = 1.0,
											double width = 0.0, rgb color = black )
		: arc( x, y, radius, 0, 360, width, color ) {}

	virtual shape*	Clone( void ) { return new circle( *this ); }
	circle&			operator +=( pvect v ) { Translate( v ); return *this; }
};

inline circle&		operator +( circle s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// wedge
class wedge : public shape {
public:
	double			mRadius;
	int 			mStart;
	int				mSweep;

	wedge( pvect center, double radius, int start, int sweep, rgb color = black )
		: shape( color, center ), mRadius( radius ), mStart( start ), mSweep( sweep ) {}
	wedge( double x, double y, double radius, int start, int sweep, rgb color = black )
		: shape( color, x, y ), mRadius( radius ), mStart( start ), mSweep( sweep ) {}

	virtual shape*	Clone( void ) { return new wedge( *this ); }
	virtual void	Draw( draw& d )
						{ d.PaintWedge( mBase, mRadius, mStart, mSweep, mColor ); }
	wedge&			operator +=( pvect v ) { Translate( v ); return *this; }
};

inline wedge&		operator +( wedge s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// disc
class disc : public wedge {
public:
	disc( pvect center = pvect( 0.0, 0.0 ), double radius = 1.0, rgb color = black )
		: wedge( center, radius, 0, 360, color ) {}
	disc( double x, double y, double radius = 1.0, rgb color = black )
		: wedge( x, y, radius, 0, 360, color ) {}

	virtual shape*	Clone( void ) { return new disc( *this ); }
	disc&			operator +=( pvect v ) { Translate( v ); return *this; }
};

inline disc&		operator +( disc s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// polygon
class polygon : public shape {
public:
	int				mNumVertices;
	int				mArraySize;
	pvect*			mVertex;

	polygon( rgb color = black, int initial_size = ARRAY_INCREMENT_SIZE );
	polygon( const polygon& p );
	~polygon();

	void			Clear( void ) { mNumVertices = 0; }
	void			AddVertex( pvect v );
	void			AddVertex( double x, double y ) { AddVertex( pvect( x, y ) ); }

	virtual shape*	Clone( void ) { return new polygon( *this ); }
	virtual void	Draw( draw& d );
	polygon&		operator +=( pvect v ) { Translate( v ); return *this; }
};

inline polygon&		operator +( polygon s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// pl_curve
class pl_curve : public polygon {
public:
	pl_curve( rgb color = black, int initial_size = ARRAY_INCREMENT_SIZE )
		: polygon( color, initial_size ) {}
	pl_curve( const pl_curve& p ) : polygon( p ) {}

	virtual shape*	Clone( void ) { return new pl_curve( *this ); }
	virtual void	Draw( draw& d );
	pl_curve&		operator +=( pvect v ) { Translate( v ); return *this; }
};

inline pl_curve&	operator +( pl_curve s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// textstring
class textstring : public shape {
public:
	string			mString;
	string			mFontName;
	double			mFontSz;
	int				mFontSize;
	int				mFixedSize;
	int				mCentering;

	textstring( pvect loc, string str, int font_size = 10, string font_name = "",
										int centering = false, rgb color = black );
	textstring( double x, double y, string str,
										int font_size = 10, string font_name = "",
										int centering = false, rgb color = black );
	textstring( pvect loc, string str, double font_sz, string font_name = "",
										int centering = false, rgb color = black );
	textstring( double x, double y, string str,
										double font_sz, string font_name = "",
										int centering = false, rgb color = black );
	textstring( textstring& s );

	virtual shape*	Clone( void ) { return new textstring( *this ); }
	virtual void	Draw( draw& d );
	textstring&		operator +=( pvect v ) { Translate( v ); return *this; }
};

inline textstring&	operator +( textstring s, pvect v ) { return ( s += v ); }


// --------------------------------------------------------------------------------
// composite : collection of shapes
class composite : public shape {
public:
	int				mNumShapes;
	int				mArraySize;
	shape**			mShape;

	composite();
	composite( const composite& s );
	~composite();

	void			Clear( void );
	void			AddShape( shape* p );
	void			AddShape( shape& s ) { AddShape( s.Clone() ); }
	composite&		operator <<( shape& s ) { AddShape( s.Clone() ); return *this; }

	void			SetColor( rgb color );

	virtual shape*	Clone( void ) { return new composite( *this ); }
	virtual void	Draw( draw& d );
	composite&		operator +=( pvect v ) { Translate( v ); return *this; }
};

inline composite&	operator +( composite s, pvect v ) { return ( s += v ); }



#endif
