// iDraw.h   1998 by Masaaki Wada. All rights reserved.
// 19980520 Masaaki Wada
// 19980610MW CWPro3

#ifndef		_iDraw_
#define		_iDraw_

#include	<string>
using namespace std;

#ifndef		ARRAY_INCREMENT_SIZE
#define		ARRAY_INCREMENT_SIZE	8
#endif



// ================================================================================
// iPoint

class iPoint {
public:
	int				i;	// horizontal coordinate; increases toward the right
	int				j;	// vertical coordinate; increases downward
	
	iPoint( int _i = 0, int _j = 0 ) { i = _i; j = _j; }

	void			Set( int _i, int _j ) { i = _i; j = _j; }

	iPoint			operator +=( iPoint p ) { i += p.i; j += p.j; return *this; }
	iPoint			operator -=( iPoint p ) { i -= p.i; j -= p.j; return *this; }
};

inline iPoint		operator +( iPoint p, iPoint q ) { return ( p += q ); }
inline iPoint		operator -( iPoint p, iPoint q ) { return ( p -= q ); }

inline int			operator ==( iPoint p, iPoint q )
						{ return ( p.i == q.i && p.j == q.j ); }
inline int			operator !=( iPoint p, iPoint q )
						{ return !( p == q ); }


// ================================================================================
// iRect

class iRect {
public:
	int				left, top, right, bottom;

	iRect( int _l = 0, int _t = 0, int _r = 1, int _b = 1 )
						{ left = _l; top = _t; right = _r; bottom = _b; }

	void			Set( int _l, int _t, int _r, int _b )
						{ left = _l; top = _t; right = _r; bottom = _b; }

	iRect			operator +=( iPoint p )
						{ left += p.i; top += p.j; right += p.i; bottom += p.j;
																		return *this; }
	iRect			operator -=( iPoint p )
						{ left -= p.i; top -= p.j; right -= p.i; bottom -= p.j;
																		return *this; }

	int				Contains( iPoint p )
						{ return ( p.i >= left && p.i <= right &&
													p.j >= top && p.j <= bottom ); }
};

inline iRect		operator +( iRect r, iPoint& p ) { return ( r += p ); }

inline int			Contains( iRect r, iPoint p ) { return r.Contains( p ); }
inline int			ContainedIn( iPoint p, iRect r ) { return r.Contains( p ); }


// ================================================================================
// rgb

enum				scale_t { color_scale, gray_scale };

class rgb {
public:
	double			red, green, blue;	// values are to be between  0.0  and  1.0

	rgb( double r = 0.0, double g = 0.0, double b = 0.0 ) { Set( r, g, b ); }
	rgb( scale_t scale, double m, double mmin = 0.0, double mmax = 1.0 );
	rgb( scale_t scale, int k, int kmin, int kmax );

	void			Set( double r, double g, double b );

	void			ColorScale( double m, double mmin = 0.0, double mmax = 1.0 );
	void			ColorScale( int k, int kmin, int kmax );
	void			GrayScale( double m, double mmin = 0.0, double mmax = 1.0 );
	void			GrayScale( int k, int kmin, int kmax );
};

const rgb			black( 0.0, 0.0, 0.0 );
const rgb			white( 1.0, 1.0, 1.0 );


// ================================================================================
// iDraw

class iDraw {

protected:	// ----------   data members   ----------
	iRect			mFrame;

public:		// ----------   constructors and accessors   ----------
	iDraw( iRect fr = iRect( 0, 0, 400, 400 ) ) : mFrame( fr ) {}
	iDraw( int _l, int _t, int _r, int _b ) : mFrame( _l, _t, _r, _b ) {}
	virtual ~iDraw() {}

	iRect			Frame( void ) { return mFrame; }
	int				Left( void ) { return mFrame.left; }
	int				Right( void ) { return mFrame.right; }
	int				Top( void ) { return mFrame.top; }
	int				Bottom( void ) { return mFrame.bottom; }
	void			SetFrame( iRect fr ) { mFrame = fr; }
	void			SetFrame( int _l, int _t, int _r, int _b )
						{ mFrame.Set( _l, _t, _r, _b ); }


public:		// ----------   drawing   ----------
// line segment of 1 pixel width
	void			DrawLine( iPoint p0, iPoint p1, rgb color = black )
						{ IDrawLine( p0.i, p0.j, p1.i, p1.j, color ); }

	void			DrawLine( int i0, int j0, int i1, int j1, rgb color = black )
						{ IDrawLine( i0, j0, i1, j1, color ); }

// rectangular frame
	void			DrawRectangle( iRect r, int width = 1, rgb color = black )
						{ IDrawRectangle( r.left, r.top, r.right, r.bottom, width, color ); }

	void			DrawRectangle( int l, int t, int r, int b, int width = 1, rgb color = black )
						{ IDrawRectangle( l, t, r, b, width, color ); }

// rectangle
	void			PaintRectangle( iRect r, rgb color = black )
						{ IPaintRectangle( r.left, r.top, r.right, r.bottom, color ); }

	void			PaintRectangle( int l, int t, int r, int b, rgb color = black )
						{ IPaintRectangle( l, t, r, b, color ); }

// oval inscribed to the specified rectangle of given width
	void			DrawOval( iRect r, int start = 0, int sweep = 360,
														int width = 1, rgb color = black )
						{ IDrawOval( r.left, r.top, r.right, r.bottom, start, sweep,
																		width, color ); }

	void			DrawOval( int l, int t, int r, int b, int start = 0, int sweep = 360,
														int width = 1, rgb color = black )
						{ IDrawOval( l, t, r, b, start, sweep, width, color ); }

// oval
	void			PaintOval( iRect r, int start = 0, int sweep = 360, rgb color = black )
						{ IPaintOval( r.left, r.top, r.right, r.bottom, start, sweep, color ); }

	void			PaintOval( int l, int t, int r, int b,
								int start = 0, int sweep = 360, rgb color = black )
						{ IPaintOval( l, t, r, b, start, sweep, color ); }

// polygon
	void			DrawPolygon( iPoint vertex[], int num_vertices, rgb color = black )
						{ IDrawPolygon( vertex, num_vertices, color ); }

	void			PaintPolygon( iPoint vertex[], int num_vertices, rgb color = black )
						{ IPaintPolygon( vertex, num_vertices, color ); }

// character string
	void			DrawString( iPoint p, string str,
												int font_size = 10, string font_name = "",
												int centering = false, rgb color = black )
						{ IDrawString( p.i, p.j, str, font_size, font_name, centering, color ); }

	void			DrawString( int i, int j, string str,
												int font_size = 10, string font_name = "",
												int centering = false, rgb color = black )
						{ IDrawString( i, j, str, font_size, font_name, centering, color ); }


protected:	// ----------   implementation: override the following   ----------
	virtual void	IDrawLine( int i0, int j0, int i1, int j1, rgb color ) {}
	virtual void	IDrawRectangle( int l, int t, int r, int b, int width, rgb color ) {}
	virtual void	IPaintRectangle( int l, int t, int r, int b, rgb color ) {}
	virtual void	IDrawOval( int l, int t, int r, int b, int start, int sweep,
																int width, rgb color ) {}
	virtual void	IPaintOval( int l, int t, int r, int b, int start, int sweep,
																			rgb color ) {}
	virtual void	IDrawPolygon( iPoint vertex[], int num_vertices, rgb color ) {}
	virtual void	IPaintPolygon( iPoint vertex[], int num_vertices, rgb color ) {}
	virtual void	IDrawString( int i, int j, string str, int font_size, string font_name,
															int centering, rgb color ) {}

};   // end of class iDraw


// ================================================================================
// shapes

// --------------------------------------------------------------------------------
// iShape : base class
class iShape {
protected:
	rgb				mColor;
	iPoint			mBase;

public:
	iShape( rgb color = black, iPoint p = iPoint( 0, 0 ) )
										: mColor( color ), mBase( p ) {}
	iShape( rgb color, int i, int j ) : mColor( color ), mBase( i, j ) {}
	virtual ~iShape() {}

	rgb				Color( void ) { return mColor; }
	virtual void	SetColor( rgb color ) { mColor = color; }
	iPoint			Base( void ) { return mBase; }
	virtual void	SetBase( iPoint p ) { mBase = p; }

	virtual iShape*	Clone( void ) { return new iShape( *this ); }
	virtual void	Draw( iDraw& d ) {}
	virtual void	Translate( iPoint p ) { mBase += p; }
	iShape&			operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iShape&		operator +( iShape s, iPoint p ) { return ( s += p ); }

inline iDraw&		operator <<( iDraw& d, iShape& s ) { s.Draw( d ); return d; }


// --------------------------------------------------------------------------------
// iPaintOut
class iPaintOut : public iShape {
public:
	iPaintOut( rgb color = white ) : iShape( color ) {}
	virtual iShape*	Clone( void ) { return new iPaintOut( *this ); }
	virtual void	Draw( iDraw& d ) { d.PaintRectangle( d.Frame(), mColor ); }
	iPaintOut&		operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iPaintOut&	operator +( iPaintOut s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iLine
class iLine : public iShape {
public:
	iPoint			mP0;
	iPoint			mP1;

	iLine( iPoint p0, iPoint p1, rgb color = black )
							: iShape( color ), mP0( p0 ), mP1( p1 ) {}
	iLine( int i0, int j0, int i1, int j1, rgb color = black )
							: iShape( color ), mP0( i0, j0 ), mP1( i1, j1 ) {}

	virtual iShape*	Clone( void ) { return new iLine( *this ); }
	virtual void	Draw( iDraw& d ) { d.DrawLine( mBase + mP0, mBase + mP1, mColor ); }
	iLine&			operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iLine&		operator +( iLine s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iRectFrame
class iRectFrame : public iShape {
public:
	iRect			mRect;
	int				mWidth;

	iRectFrame( iRect r, int width = 1, rgb color = black )
							: iShape( color ), mRect( r ), mWidth( width ) {}
	iRectFrame( int l, int t, int r, int b, int width = 1, rgb color = black )
							: iShape( color ), mRect( l, t, r, b ), mWidth( width ) {}

	virtual iShape*	Clone( void ) { return new iRectFrame( *this ); }
	virtual void	Draw( iDraw& d ) { d.DrawRectangle( mRect + mBase, mWidth, mColor ); }
	iRectFrame&		operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iRectFrame&	operator +( iRectFrame s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iRectangle
class iRectangle : public iShape {
public:
	iRect			mRect;

	iRectangle( iRect r, rgb color = black ) : iShape( color ), mRect( r ) {}
	iRectangle( int l, int t, int r, int b, rgb color = black )
							: iShape( color ), mRect( l, t, r, b ) {}

	virtual iShape*	Clone( void ) { return new iRectangle( *this ); }
	virtual void	Draw( iDraw& d ) { d.PaintRectangle( mRect + mBase, mColor ); }
	iRectangle&		operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iRectangle&	operator +( iRectangle s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iOvalFrame
class iOvalFrame : public iShape {
public:
	iRect			mRect;
	int				mStart;
	int				mSweep;
	int				mWidth;

	iOvalFrame( iRect r, int start = 0, int sweep = 360, int width = 1, rgb color = black )
							: iShape( color ), mRect( r ),
								mStart( start ), mSweep( sweep ), mWidth( width ) {}
	iOvalFrame( int l, int t, int r, int b, int start = 0, int sweep = 360,
														int width = 1, rgb color = black )
							: iShape( color ), mRect( l, t, r, b ),
								mStart( start ), mSweep( sweep ), mWidth( width ) {}

	virtual iShape*	Clone( void ) { return new iOvalFrame( *this ); }
	virtual void	Draw( iDraw& d )
						{ d.DrawOval( mRect + mBase, mStart, mSweep, mWidth, mColor ); }
	iOvalFrame&		operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iOvalFrame&	operator +( iOvalFrame s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iOval
class iOval : public iShape {
public:
	iRect			mRect;
	int				mStart;
	int				mSweep;

	iOval( iRect r, int start = 0, int sweep = 360, rgb color = black )
							: iShape( color ), mRect( r ),
								mStart( start ), mSweep( sweep ) {}
	iOval( int l, int t, int r, int b, int start = 0, int sweep = 360, rgb color = black )
							: iShape( color ), mRect( l, t, r, b ),
								mStart( start ), mSweep( sweep ) {}

	virtual iShape*	Clone( void ) { return new iOval( *this ); }
	virtual void	Draw( iDraw& d )
						{ d.PaintOval( mRect + mBase, mStart, mSweep, mColor ); }
	iOval&			operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iOval&		operator +( iOval s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iPolyFrame
class iPolyFrame : public iShape {
public:
	int				mNumVertices;
	int				mArraySize;
	iPoint*			mVertex;

	iPolyFrame( rgb color = black, int initial_size = ARRAY_INCREMENT_SIZE );
	iPolyFrame( const iPolyFrame& poly );
	~iPolyFrame();

	void			Clear( void ) { mNumVertices = 0; }
	void			AddVertex( iPoint p );
	void			AddVertex( int i, int j ) { AddVertex( iPoint( i, j ) ); }
	iPolyFrame&		operator <<( iPoint p ) { AddVertex( p ); return *this; }

	virtual iShape*	Clone( void ) { return new iPolyFrame( *this ); }
	virtual void	Draw( iDraw& d );
	iPolyFrame&		operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iPolyFrame&	operator +( iPolyFrame s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iPolygon
class iPolygon : public iShape {
public:
	int				mNumVertices;
	int				mArraySize;
	iPoint*			mVertex;

	iPolygon( rgb color = black, int initial_size = ARRAY_INCREMENT_SIZE );
	iPolygon( const iPolygon& poly );
	~iPolygon();

	void			Clear( void ) { mNumVertices = 0; }
	void			AddVertex( iPoint p );
	void			AddVertex( int i, int j ) { AddVertex( iPoint( i, j ) ); }
	iPolygon&		operator <<( iPoint p ) { AddVertex( p ); return *this; }

	virtual iShape*	Clone( void ) { return new iPolygon( *this ); }
	virtual void	Draw( iDraw& d );
	iPolygon&		operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iPolygon&	operator +( iPolygon s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iTextString
class iTextString : public iShape {
public:
	string			mString;
	int				mFontSize;
	string			mFontName;
	int				mCentering;

	iTextString( iPoint p, string str, int font_size = 10, string font_name = "",
										int centering = false, rgb color = black )
							: iShape( color, p )
						{
							mString = str;
							mFontSize = font_size;
							mFontName = font_name;
							mCentering = centering;
						}
	iTextString( int i, int j, string str, int font_size = 10, string font_name = "",
										int centering = false, rgb color = black )
							: iShape( color, i, j )
						{
							mString = str;
							mFontSize = font_size;
							mFontName = font_name;
							mCentering = centering;
						}
	iTextString( const iTextString& s ) : iShape( s.mColor, s.mBase )
						{
							mString = s.mString;
							mFontSize = s.mFontSize;
							mFontName = s.mFontName;
							mCentering = s.mCentering;
						}
	virtual iShape*	Clone( void ) { return new iTextString( *this ); }
	virtual void	Draw( iDraw& d )
						{ d.DrawString( mBase, mString, mFontSize, mFontName,
															mCentering, mColor ); }
	iTextString&	operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iTextString&	operator +( iTextString s, iPoint p ) { return ( s += p ); }


// --------------------------------------------------------------------------------
// iComposite : collection of iShapes
class iComposite : public iShape {
public:
	int				mNumShapes;
	int				mArraySize;
	iShape**		mShape;

	iComposite();
	iComposite( const iComposite& s );
	~iComposite();

	void			Clear( void );
	void			AddShape( iShape* p );
	void			AddShape( iShape& s ) { AddShape( s.Clone() ); }
	iComposite&		operator <<( iShape& s ) { AddShape( s.Clone() ); return *this; }

	void			SetColor( rgb color );

	virtual iShape*	Clone( void ) { return new iComposite( *this ); }
	virtual void	Draw( iDraw& d );
	iComposite&		operator +=( iPoint p ) { Translate( p ); return *this; }
};

inline iComposite&	operator +( iComposite s, iPoint p ) { return ( s += p ); }


#endif
